@@ -19,10 +19,3 @@ pub async fn get_region_count() -> Result<u64> {
         .as_u64()
         .ok_or_else(|| Error::StringError("pd region count does not return an integer".to_owned()))
 }
-
-#[tokio::test]
-#[ignore]
-async fn test_get_region_count() -> Result<()> {
-    println!("{}", get_region_count().await?);
-    Ok(())
-}
